// Copyright (c) 2014, The Shelf Bind project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

library shelf_bind.bindings.param;

import 'package:shelf/shelf.dart' show Request;
import 'package:shelf_path/shelf_path.dart'
    show getPathParameter, getPathParameters;

import 'dart:mirrors'
    show
        ClosureMirror,
        InstanceMirror,
        MirrorSystem,
        ParameterMirror,
        TypeMirror,
        reflect,
        reflectClass;
import '../converters.dart' show convertTo;
import '../preconditions.dart' show ensure, isNotNull;
import '../annotations.dart' show BodyFormat, RequestBody, ContentType;
import 'dart:async' show Future;

import 'type_bindings.dart' show BindMode, TypeBinding, inferTypeBinding;
import 'dart:convert' show JSON;
import 'mirror_util.dart' show getAnnotation;

final ParameterBindingInferencer _defaultParameterInferencer =
    new ParameterBindingInferencer();

Map _mapValues(Map map, convert(v)) {
  final result = {};
  map.forEach((k, v) => result[k] = convert(v));
  return result;
}

List<HandlerParameterBinding> inferHandlerParameterBindings(
    ClosureMirror functionMirror,
    List<HandlerParameterBinding> providedBindings,
    Map<Type, CustomObjectFactory> customObjects) {
  final bindingInferencer = customObjects == null
      ? _defaultParameterInferencer
      : new ParameterBindingInferencer(
          additionalBindingFactories: _mapValues(
              customObjects,
              (factory) => (String name) =>
                  new CustomObjectParameterBinding(name, factory)));

  return bindingInferencer.inferHandlerParameterBindings(
      functionMirror, providedBindings);
}

/**
 * A binding from instances of [Request] to a single parameter of a handler
 * function.
 *
 * Different sorts of bindings are supported such as:
 * * the [request] object itself (see [RequestObjectParameterBinding])
 * * path parameters to simple type (see [SimplePathParameterBinding])
 * * path parameters to class types (see [TypePathParameterBinding])
 * * request body in JSON format to class types (see [JsonBodyTypeParameterBinding])
 */
abstract class HandlerParameterBinding {
  final String handlerParameterName;

  const HandlerParameterBinding(this.handlerParameterName);
  dynamic createInstance(Request request);
  void validate();
}

typedef CustomObjectFactory(Request request);

class CustomObjectParameterBinding extends HandlerParameterBinding {
  final CustomObjectFactory factory;

  const CustomObjectParameterBinding(String handlerParameterName, this.factory)
      : super(handlerParameterName);

  @override
  void validate() {}

  @override
  createInstance(Request request) => factory(request);
}

Request _noopCustomObjectFactory(request) => request;

class RequestObjectParameterBinding extends CustomObjectParameterBinding {
  RequestObjectParameterBinding(String handlerParameterName)
      : super(handlerParameterName, _noopCustomObjectFactory);
}

class SimplePathParameterBinding extends HandlerParameterBinding {
  final String pathParameterName;
  final Type simpleType;

  SimplePathParameterBinding(String pathParameterName,
      [this.simpleType = String])
      : this.pathParameterName = pathParameterName,
        super(pathParameterName);

  @override
  dynamic createInstance(Request request) {
    // TODO: simple type conversion
    final value = getPathParameter(request, pathParameterName);
    return value == null || simpleType == String
        ? value
        : convertTo(simpleType, value);
  }

  @override
  void validate() {
    ensure(pathParameterName, isNotNull);
  }
}

abstract class _TypeParameterBinding extends HandlerParameterBinding {
  final TypeBinding typeBinding;

  _TypeParameterBinding(String handlerParameterName, this.typeBinding)
      : super(handlerParameterName);

  @override
  void validate() {
    ensure(typeBinding, isNotNull);
    typeBinding.validateBindings();
  }

  @override
  createInstance(Request request) => _createUnvalidated(request);

  _createUnvalidated(Request request);

  _TypeParameterBinding copy(TypeBinding newTypeBinding);
}

class TypePathParameterBinding extends _TypeParameterBinding {
  TypePathParameterBinding(String handlerParameterName, TypeBinding typeBinding)
      : super(handlerParameterName, typeBinding);

  @override
  dynamic _createUnvalidated(Request request) {
    final instanceMirror =
        _createInstance(getPathParameters(request), typeBinding);
    return instanceMirror.reflectee;
  }

  @override
  _TypeParameterBinding copy(TypeBinding newTypeBinding) =>
      new TypePathParameterBinding(handlerParameterName, newTypeBinding);
}

abstract class BodyParameterBinding extends _TypeParameterBinding {
  BodyParameterBinding(String handlerParameterName, TypeBinding typeBinding)
      : super(handlerParameterName, typeBinding);
}

class JsonBodyTypeParameterBinding extends BodyParameterBinding {
  JsonBodyTypeParameterBinding(
      String handlerParameterName, TypeBinding typeBinding)
      : super(handlerParameterName, typeBinding);

  @override
  dynamic _createUnvalidated(Request request) {
    return _extractJsonBody(request).then((json) {
      final classMirror = reflectClass(typeBinding.type);
      final instanceMirror = classMirror.newInstance(#fromJson, [json]);

      return instanceMirror.reflectee;
    });
  }

  Future<Map> _extractJsonBody(Request request) =>
      request.readAsString().then((s) => JSON.decode(s));

  @override
  _TypeParameterBinding copy(TypeBinding newTypeBinding) =>
      new JsonBodyTypeParameterBinding(handlerParameterName, newTypeBinding);
}

class FormBodyTypeParameterBinding extends BodyParameterBinding {
  FormBodyTypeParameterBinding(
      String handlerParameterName, TypeBinding typeBinding)
      : super(handlerParameterName, typeBinding);

  @override
  dynamic _createUnvalidated(Request request) {
    return _extractFormBody(request).then((formParams) {
      final instanceMirror = _createInstance(formParams, typeBinding);
      return instanceMirror.reflectee;
    });
  }

  Future<Map> _extractFormBody(Request request) =>
      request.readAsString().then((s) => Uri.splitQueryString(s));

  @override
  _TypeParameterBinding copy(TypeBinding newTypeBinding) =>
      new FormBodyTypeParameterBinding(handlerParameterName, newTypeBinding);
}

InstanceMirror _createInstance(
    Map<String, Object> parameters, TypeBinding binding) {
  switch (binding.bindMode) {
    case BindMode.CONSTRUCTOR:
      return _createInstanceViaConstructor(parameters, binding);
    case BindMode.PROPERTY:
      return _createInstanceViaProperty(parameters, binding);

    default:
      throw new StateError('unsupported bind mode: ${binding.bindMode}');
  }
}

InstanceMirror _createInstanceViaConstructor(
    Map<String, Object> parameters, TypeBinding binding) {
  final namedArguments = _bindVariables(parameters, binding.memberBindings);
  final classMirror = reflectClass(binding.type);
  final instanceMirror =
      classMirror.newInstance(binding.constructorName, [], namedArguments);
  return instanceMirror;
}

InstanceMirror _createInstanceViaProperty(
    Map<String, Object> parameters, TypeBinding binding) {
  final properties = _bindVariables(parameters, binding.memberBindings);
  final classMirror = reflectClass(binding.type);
  final instanceMirror = classMirror.newInstance(const Symbol(''), []);

  properties.forEach((property, value) {
    instanceMirror.setField(property, value);
  });
  return instanceMirror;
}

Map<Symbol, dynamic> _bindVariables(
    Map<String, Object> parameters, Map<Symbol, Symbol> memberBindings) {
  final convertedPv = <Symbol, Object>{};
  parameters.forEach((k, v) {
    convertedPv[new Symbol(k)] = v;
  });

  final namedArguments = <Symbol, dynamic>{};
  memberBindings.forEach((k, v) {
    final value = convertedPv[k];
    if (value != null) {
      namedArguments[v] = value;
    }
  });

  return namedArguments;
}

typedef HandlerParameterBinding HandlerParameterBindingFactory(String name);

class ParameterBindingInferencer {
  static final Map<Type,
      HandlerParameterBindingFactory> _defaultBindingFactories = {
    Request: (String name) => new RequestObjectParameterBinding(name)
  };

  final Map<Type, HandlerParameterBindingFactory> bindingFactories;

  ParameterBindingInferencer(
      {Map<Type, HandlerParameterBindingFactory> additionalBindingFactories})
      : this.bindingFactories = additionalBindingFactories != null
            ? ({}
              ..addAll(additionalBindingFactories)
              ..addAll(_defaultBindingFactories))
            : _defaultBindingFactories;

  List<HandlerParameterBinding> inferHandlerParameterBindings(
      ClosureMirror functionMirror,
      List<HandlerParameterBinding> providedBindings) {
    final params = functionMirror.function.parameters;
    final inferredBindings = <HandlerParameterBinding>[];

    for (int i = 0; i < params.length; i++) {
      final param = params[i];
      final providedBinding = i > providedBindings.length - 1
          ? null
          : providedBindings.elementAt(i);
      final parameterBinding =
          _inferHandlerParameterBinding(providedBinding, param);
      if (parameterBinding != null) {
        inferredBindings.add(parameterBinding);
      }
    }
    return inferredBindings;
  }

  HandlerParameterBinding _inferHandlerParameterBinding(
      HandlerParameterBinding providedBinding, ParameterMirror param) {
    final functionParameterName = MirrorSystem.getName(param.simpleName);

    // TODO: better name than defaultBinding
    final defaultBindingFactory = bindingFactories[param.type.reflectedType];
    if (defaultBindingFactory != null) {
      return defaultBindingFactory(functionParameterName);
    }

    if (_isSimpleType(param.type)) {
      return _inferSimplePathParameterBinding(
          providedBinding, param, functionParameterName);
    } else {
      return _inferTypeParameterBinding(
          providedBinding, param, functionParameterName);
    }
  }

  HandlerParameterBinding _inferSimplePathParameterBinding(
      HandlerParameterBinding providedBinding,
      ParameterMirror param,
      String functionParameterName) {
    if (providedBinding != null && providedBinding is _TypeParameterBinding) {
      throw new ArgumentError(
          'type bindings not supported for dart core classes');
    }

    final providedName = providedBinding == null
        ? null
        : (providedBinding as SimplePathParameterBinding).pathParameterName;

    final pathParameterName =
        providedName != null ? providedName : functionParameterName;

    return new SimplePathParameterBinding(
        pathParameterName, param.type.reflectedType);
  }

  HandlerParameterBinding _inferTypeParameterBinding(
      HandlerParameterBinding providedBinding,
      ParameterMirror param,
      String functionParameterName) {
    if (providedBinding != null &&
        providedBinding is SimplePathParameterBinding) {
      throw new ArgumentError(
          'simple bindings not supported for non dart core classes');
    }

    providedBinding = _inferTypeParameterBindingFromAnnotation(
        providedBinding, param, functionParameterName);

    final providedTypeBinding = providedBinding == null
        ? new TypeBinding(null)
        : (providedBinding as _TypeParameterBinding).typeBinding;

    final inferredTypeBinding = inferTypeBinding(providedTypeBinding, param);

    final inferredParameterBinding = providedBinding == null
        ? new TypePathParameterBinding(
            functionParameterName, inferredTypeBinding)
        : (providedBinding as _TypeParameterBinding).copy(inferredTypeBinding);

    inferredParameterBinding.validate();
    return inferredParameterBinding;
  }

  HandlerParameterBinding _inferTypeParameterBindingFromAnnotation(
      HandlerParameterBinding providedBinding,
      ParameterMirror param,
      String functionParameterName) {
    final RequestBody requestBodyAnnotation = getAnnotation(param, RequestBody);

    if (requestBodyAnnotation == null) {
      return providedBinding;
    }

    final TypeBinding typeBinding = new TypeBinding(requestBodyAnnotation.type);

    switch (requestBodyAnnotation.format) {
      case ContentType.JSON:
        return new JsonBodyTypeParameterBinding(
            functionParameterName, typeBinding);
      case ContentType.FORM:
        return new FormBodyTypeParameterBinding(
            functionParameterName, typeBinding);
      default:
        throw new ArgumentError(
            "Only JSON and FORM bodies currently supported");
    }
  }

// TODO: Not a good way to determine this
  bool _isSimpleType(TypeMirror tm) =>
      (tm.isTopLevel && tm.owner.simpleName == #dart.core);
}
